# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import shutil
import warnings

import numpy as np


# Intended as a fixture. Used in conftest.py.
def temp_dir(request):

    tmpfs = os.path.normpath(os.path.join('/', 'dev', 'shm'))
    # would like to check mount type, but that requires system calls
    if os.path.exists(tmpfs) and os.path.ismount(tmpfs):

        base_path = tmpfs

    else:

        base_path = os.path.dirname(__file__)

    fls = os.listdir(base_path)
    while True:
        dname = ''.join(map(str, np.random.randint(0, 10, 6)))
        if "allensdk_test_{}".format(dname) not in fls:
            break

    specific_path = os.path.join(base_path, 'allensdk_test_' + dname)
    os.makedirs(specific_path)

    def fin():
        shutil.rmtree(specific_path)
        if os.path.exists(specific_path):
            warnings.warn('test dir {0} still exists!', UserWarning)

    request.addfinalizer(fin)

    return specific_path



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os

import pandas as pd
import pandas.io.json as pj
import numpy as np

import pytest
from mock import MagicMock, mock_open, patch

from allensdk.api.cache import Cache, memoize, get_default_manifest_file
from allensdk.api.queries.rma_api import RmaApi
import allensdk.core.json_utilities as ju
from allensdk.config.manifest import ManifestVersionError
from allensdk.config.manifest_builder import ManifestBuilder

_msg = [{'whatever': True}]
_pd_msg = pd.DataFrame(_msg)


@pytest.fixture
def cache():
    return Cache()


@pytest.fixture
def rma():
    return RmaApi()


@pytest.fixture
def wavefront_obj():
    return '''

v 8578 5484.96 5227.57
v 8509.2 5487.54 5237.07
v 8564.38 5522.13 5220.41
v 8631.93 5497.82 5228.33
v 8517.88 5542.95 5234.53
v 8615.26 5563.22 5224.48

# i'm a comment!

vn -0.0247061 -0.352726 -0.935401
vn -0.235489 -0.190095 -0.953105
vn -0.0880336 -0.0323767 -0.995591
vn 0.122706 -0.209891 -0.969994
vn -0.343738 0.217978 -0.913416
vn 0.0753706 0.16324 -0.983703

I should be a comment, but am not

f 1//1 2//2 3//3
f 4//4 1//1 3//3
f 3//3 2//2 5//5
f 6//6 3//3 5//5

    '''


@pytest.fixture
def dummy_cache():
    class DummyCache(Cache):

        VERSION = None

        def build_manifest(self, file_name):
            manifest_builder = ManifestBuilder()
            manifest_builder.set_version(DummyCache.VERSION)
            manifest_builder.write_json_file(file_name)

    return DummyCache


def test_version_update(fn_temp_dir, dummy_cache):

    mpath = os.path.join(fn_temp_dir, 'manifest.json')
    dc = dummy_cache(manifest=mpath)

    same_dc = dummy_cache(manifest=mpath)

    with pytest.raises(ManifestVersionError):
        new_dc = dummy_cache(manifest=mpath, version=1.0)


def test_load_manifest(tmpdir_factory, dummy_cache):

    manifest = tmpdir_factory.mktemp('data').join('test_manifest.json')
    cache = dummy_cache(manifest=str(manifest))

    assert(cache.manifest_path == str(manifest))
    assert(os.path.exists(cache.manifest_path))


@patch("allensdk.core.json_utilities.write")
@patch("allensdk.core.json_utilities.read", return_value=pd.DataFrame(_msg))
@patch("allensdk.core.json_utilities.read_url_get", return_value={'msg': _msg})
def test_wrap_json(ju_read_url_get, ju_read, ju_write, rma, cache):
    df = cache.wrap(rma.model_query,
                    'example.txt',
                    cache=True,
                    model='Hemisphere')

    assert df.loc[:, 'whatever'][0]

    ju_read_url_get.assert_called_once_with(
        'http://api.brain-map.org/api/v2/data/query.json?q=model::Hemisphere')
    ju_write.assert_called_once_with('example.txt', _msg)
    ju_read.assert_called_once_with('example.txt')


@patch("pandas.io.json.read_json", return_value=_msg)
@patch("allensdk.core.json_utilities.write")
@patch("allensdk.core.json_utilities.read_url_get", return_value={'msg': _msg})
def test_wrap_dataframe(ju_read_url_get, ju_write, mock_read_json, rma, cache):
    json_data = cache.wrap(rma.model_query,
                           'example.txt',
                           cache=True,
                           return_dataframe=True,
                           model='Hemisphere')

    assert json_data[0]['whatever']

    ju_read_url_get.assert_called_once_with(
        'http://api.brain-map.org/api/v2/data/query.json?q=model::Hemisphere')
    ju_write.assert_called_once_with('example.txt', _msg)
    mock_read_json.assert_called_once_with('example.txt', orient='records')

def test_memoize():

        import time

        @memoize
        def f(x):
            time.sleep(1)
            return x

        for ii in range(2):
            t0 = time.time()
            print(f(0), time.time() - t0)

        class FooBar(object):

            def __init__(self): pass

            @memoize
            def f(self, x):
                time.sleep(.1)
                return 1

        fb = FooBar()

        for ii in range(2):
            t0 = time.time()
            fb.f(0), time.time() - t0


def test_get_default_manifest_file():
    assert get_default_manifest_file('brain_observatory') == 'brain_observatory/manifest.json'
    assert get_default_manifest_file('cell_types') == 'cell_types/manifest.json'
    assert get_default_manifest_file('mouse_connectivity') == 'mouse_connectivity/manifest.json'



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2018. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import pytest
import numpy as np
import SimpleITK as sitk
import nrrd

from allensdk.core import sitk_utilities as su



@pytest.fixture(params=[1, 2, 3, 4])
def ncomponents(request):
    return request.param


@pytest.fixture(params=[ [10, 20], [10, 20, 30], [10, 20, 30], [10, 10, 10], [20, 20] ])
def size(request):
    return request.param


@pytest.fixture(params=[ lambda x: list(range(x+1))[1:], lambda x: [10] * x ])
def spacing(request):
    return request.param


@pytest.fixture(params=[ lambda x: list(range(x+1))[1:], lambda x: [5] * x ])
def origin(request):
    return request.param


@pytest.fixture(params=[ lambda x: np.eye(x).flatten() ])
def direction(request):
    return request.param


@pytest.fixture(scope='function')
def image(size, ncomponents, spacing, origin, direction):

    if ncomponents > 1:
        img = sitk.Image(size, sitk.sitkVectorUInt8, ncomponents)
    else:
        img = sitk.Image(size, sitk.sitkUInt8, ncomponents)

    ndim = len(size)

    spacing_val = spacing(ndim)
    img.SetSpacing(spacing_val)

    origin_val = origin(ndim)
    img.SetOrigin(origin_val)

    dir_val = direction(ndim)
    img.SetDirection(dir_val)

    return img, {'ncomponents': ncomponents, 
                 'size': size,
                 'spacing': spacing_val, 
                 'origin': origin_val, 
                 'direction': dir_val}


def test_get_sitk_image_information(image):

    obtained = su.get_sitk_image_information(image[0])
    for key, value in image[1].items():
        assert(np.allclose( obtained[key], value ))


def test_set_sitk_image_information_roundtrip(image):

    info = su.get_sitk_image_information(image[0])
    arr = sitk.GetArrayFromImage(image[0])

    new_image = sitk.GetImageFromArray(arr, info['ncomponents'] > 1)
    su.set_sitk_image_information(new_image, info)

    obtained = su.get_sitk_image_information(new_image)
    for key, value in info.items():
        assert(np.allclose( obtained[key], value ))


@pytest.mark.parametrize('act,dec,nc', [ ([10, 20], [20, 10], 1), 
                                         ([10, 20, 30], [30, 20, 10], 1),
                                         ([10, 20, 30, 3], [30, 20, 10], 3),
                                         ([10, 10, 10, 3], [10, 10, 10], 3 ) ])
def test_fix_array_dimensions(act, dec, nc):

    arr = np.zeros(act)
    obt = su.fix_array_dimensions(arr, nc)

    if nc == 1:
        assert(np.array_equal( obt.shape, dec ))
    else:
        assert(np.array_equal( obt.shape[:-1], dec ))

    assert( not np.isfortran(obt) )


def test_sitk_metaimage_roundtrip(tmpdir_factory, size):

    path = tmpdir_factory.mktemp('metaimage_io_test').join('dummy.mhd')
    
    array = np.random.rand(*size)

    su.write_ndarray_with_sitk(array, path)
    obt_image, obt_info = su.read_ndarray_with_sitk(path)

    assert(np.allclose( obt_image, array ))


def test_sitk_metaimage_vector_roundtrip(tmpdir_factory, size):

    path = tmpdir_factory.mktemp('metaimage_io_test').join('dummy.mhd')
    
    size = list(size) + [3]
    array = np.random.rand(*size)

    su.write_ndarray_with_sitk(array, path, ncomponents=3)
    obt_image, obt_info = su.read_ndarray_with_sitk(path)

    assert(np.allclose( obt_image, array ))
    assert( obt_info['ncomponents'] == 3 )


def test_sitk_nrrd_read(tmpdir_factory, size):
    
    path = tmpdir_factory.mktemp('nrrd_io_test').join('dummy.nrrd')
    array = np.random.rand(*size)
    
    nrrd.write(str(path), array)

    obt_image, obt_info = su.read_ndarray_with_sitk(path)

    assert(np.allclose( obt_image, array ))



def test_sitk_nrrd_write(tmpdir_factory, size):

    path = tmpdir_factory.mktemp('nrrd_io_test').join('dummy_again.nrrd')
    
    array = np.random.rand(*size)

    su.write_ndarray_with_sitk(array, path)
    obt_image, obt_info = nrrd.read(str(path))

    assert(np.allclose( obt_image, array ))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import os

import pytest

from scipy.ndimage.interpolation import zoom
import scipy.stats as stats
import numpy as np

from allensdk.brain_observatory.receptive_field_analysis import chisquarerf as chi


@pytest.fixture
def rf_events():
    
    np.random.seed(12)

    def make(receptive_field_mask, lsn):
        activity = np.logical_or(lsn == 255, lsn==0)
        return np.logical_and(activity, receptive_field_mask).sum(axis=(1, 2))[:, None]

    return make


@pytest.fixture
def locally_sparse_noise():

    def make(ntr, nr, nc):
        return np.around(np.random.rand(ntr, nr, nc)*255).astype(int)

    return make


@pytest.fixture
def rf_mask():

    def make(nr, nc, slices):
        mask = np.zeros((nr, nc))
        mask[slices] = 1
        return mask
  
    return make


@pytest.fixture
def exclusion_mask():
    mask = np.zeros((4, 4, 2))
    mask[:, :2, :] = 1
    return mask


@pytest.fixture
def events_per_pixel():

    epp = np.zeros((2, 4, 4, 2))
    epp[0, 0, 0, 0] = 2
    epp[0, 0, 0, 1] = 3
    epp[1, 3, 3, 0] = 5
    epp[1, 1, 0, 0] = 4
  
    return epp


@pytest.fixture
def trials_per_pixel():

    tpp = np.zeros((4, 4, 2))
    tpp[:, :, 0] = 2
    tpp[:, :, 1] = 0

    return tpp


# not testing d < 1 here
@pytest.mark.parametrize('r,c,d', [[2, 3, 4], [28, 16, 3], [28, 16, 2], [10, 20, 12]])
def test_interpolate_rf(r, c, d):

    image = np.arange( r * c ).reshape([ r, c ])

    delta_col = 1.0 / d
    delta_row = c * delta_col

    obtained = chi.interpolate_RF(image, d)
    grad = np.gradient(obtained)

    assert(np.allclose( grad[0], np.zeros_like(grad[0]) + delta_row ))
    assert(np.allclose( grad[1], np.zeros_like(grad[1]) + delta_col ))


# tests integration with interpolate
# not testing case where r, c are small
@pytest.mark.parametrize('r,c,d', [[28, 16, 3], [28, 16, 2], [10, 20, 12]])
def test_deinterpolate_rf(r, c, d):

    image = np.arange( r * c ).reshape([ r, c ])

    interp = chi.interpolate_RF(image, d)
    obt = chi.deinterpolate_RF(interp, c, r, d)

    assert(np.allclose( image, obt ))


def test_smooth_sta():

    image = np.eye(10)

    smoothed = chi.smooth_STA(image)
    
    thresholded = smoothed.copy()
    thresholded[thresholded < 0.5] = 0
    thresholded[thresholded > 0.5] = 1

    assert(np.allclose( smoothed.T, smoothed ))
    assert(np.allclose( image, thresholded ))
    assert( np.count_nonzero(smoothed) > np.count_nonzero(image) )


def test_build_trial_matrix():

    tr0 = np.eye(16) * 255
    tr1 = np.arange(256).reshape((16, 16))
    lsn_template = np.array([ tr0, tr1 ])

    exp = np.zeros((16, 16, 2, 2))
    exp[:, :, 0, 0] = np.eye(16)
    exp[:, :, 1, 0] = 1 - np.eye(16)
    exp[15, 15, 0, 1] = 1
    exp[0, 0, 1, 1] = 1

    obt = chi.build_trial_matrix( lsn_template, 2 )
    assert(np.allclose( exp, obt ))


def test_get_expected_events_by_pixel(exclusion_mask, events_per_pixel, trials_per_pixel):

    obt = chi.get_expected_events_by_pixel(exclusion_mask, events_per_pixel, trials_per_pixel)

    assert( obt[0, 0, 0, 0] == 0.625 ) # 5 events, 8 trials (events counted even if 0 trials)
    assert( obt[1, 1, 0, 0] == 0.5 ) # 4 events, 8 trials
    assert( obt[0, 0, 0, 1] == 0.0 ) # no trials
    assert( obt[1, 3, 3, 0] == 0.0 ) # out of mask


def test_chi_square_within_mask(exclusion_mask, events_per_pixel, trials_per_pixel):

    obt_p, obt_ch = chi.chi_square_within_mask(exclusion_mask, events_per_pixel, trials_per_pixel)

    resps = np.array([4, 0, 0, 0, 0, 0, 0, 0])
    resids = resps - 0.5
    chi_sum = (resids ** 2 / 0.5).sum()

    exp_p = 1.0 - stats.chi2.cdf(chi_sum, 15)

    # the zeroth test cell has a response without a trial.
    # this is infinitely surprising, so the pval is 0
    assert(np.allclose( obt_p, [0, exp_p] )) 


def test_get_disc_masks():

    lsn_template = np.zeros((9, 3, 3)) + 128
    for ii in range(3):
        for jj in range(3):
            lsn_template[3*ii+jj, ii, jj] = 0
    lsn_template[4, 2, 2] = 255

    exp1 = np.ones((3, 3))
    exp1[2, 2] = 0

    exp0 = np.zeros((3, 3))
    exp0[:2, :2] = 1

    obt = chi.get_disc_masks(lsn_template, radius=1)

    assert(np.allclose( exp1, obt[1, 1, :, :] ))
    assert(np.allclose( exp0, obt[0, 0, :, :] ))


def test_get_events_per_pixel():
    
    events = np.zeros((3, 2))
    trials = np.zeros((4, 4, 2, 3))

    # pixel 1,1 is off trial 1 and on trial 2 
    trials[1, 1, 1, 1] = 1
    trials[1, 1, 0, 2] = 1

    # pixel 2,2 is on trial 2 and off trial 0
    trials[2, 2, 0, 2] = 1
    trials[2, 2, 1, 0] = 1

    # cell 0 has 4 events on trial 2 and 1 on trial 0
    events[2, 0] = 4
    events[0, 0] = 1
  
    # cell 1 has 2 events on trial 1
    events[1, 1] = 2

    exp = np.zeros((2, 4, 4, 2))
    exp[0, 2, 2, 0] = 4
    exp[0, 1, 1, 0] = 4
    exp[0, 2, 2, 1] = 1
    exp[1, 1, 1, 1] = 2

    obt = chi.get_events_per_pixel(events, trials)
    assert(np.allclose( obt, exp ))


@pytest.mark.parametrize('base,ex', [[5., 10], [0.1, 12], [np.arange(20), np.linspace(0, 1, 20)]])
def test_nll_to_pvalue(base, ex):

    obt = chi.NLL_to_pvalue(ex, base)
    exp = np.power(base, -ex)

    assert(np.allclose( exp, obt ))


# test by reversing nll_to_pvalue
@pytest.mark.parametrize('base,ex', [[10., 2], [10., 4], [np.array([10, 10, 10]), np.linspace(0, 1, 3)]])
def test_pvalue_to_nll(base, ex):

    pv = chi.NLL_to_pvalue(ex, base)
    max_nll = np.amax(ex)

    obt = chi.pvalue_to_NLL(pv, max_nll)

    assert(np.allclose( ex, obt ))


@pytest.mark.skipif(os.getenv('NO_TEST_RANDOM') == 'true', reason="random seed may not produce the same results on all machines")
def test_chi_square_binary(locally_sparse_noise, rf_events, rf_mask):

    ntr = 2000
    nr = 20
    nc = 20
    slices = [slice(9, 11), slice(9, 11)]

    mask = rf_mask(nr, nc, slices)
    lsn = locally_sparse_noise(ntr, nr, nc)
    events = rf_events(mask, lsn)

    obt = chi.chi_square_binary(events, lsn)
    assert( obt[0][slices].sum() == 0 )
    assert( obt.sum() > 0 )


@pytest.mark.skipif(os.getenv('NO_TEST_RANDOM') == 'true', reason="random seed may not produce the same results on all machines")
def test_get_peak_significance(locally_sparse_noise, rf_events, rf_mask):

    ntr = 2000
    nr = 20
    nc = 20
    slices = [slice(9, 11), slice(9, 11)]

    mask = rf_mask(nr, nc, slices)
    lsn = locally_sparse_noise(ntr, nr, nc)
    events = rf_events(mask, lsn)

    chi_pv = chi.chi_square_binary(events, lsn)
    chi_nll = chi.pvalue_to_NLL(chi_pv)

    significant_cells, best_p, _, _ = chi.get_peak_significance(chi_nll, lsn)

    assert(np.allclose( best_p, 0 ))
    assert(np.allclose( significant_cells, [True] ))


def test_locate_median():

    mask = np.eye(9)
    where = np.where(mask)
    
    obt = chi.locate_median(*where)
    assert(np.allclose( obt , [4, 4] ))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2018. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import warnings

import SimpleITK as sitk
import numpy as np


def get_sitk_image_information(image):
    ''' Extract information about a SimpleITK image

    Parameters
    ----------
    image : sitk.Image
        Extract information about this image.

    Returns
    -------
    dict : 
        Extracted information. Includes spacing, origin, size, direction, and 
        number of components per pixel

    '''

    return {'spacing': image.GetSpacing(),
            'origin': image.GetOrigin(),
            'size': image.GetSize(),
            'direction': image.GetDirection(),
            'ncomponents': image.GetNumberOfComponentsPerPixel()}


def set_sitk_image_information(image, information):
    ''' Set information on a SimpleITK image

    Parameters
    ----------
    image : sitk.Image
        Set information on this image.
    information : dict
        Stores information to be set. Supports spacing, origin, direction. Also 
        checks (but cannot set) size and number of components per pixel

    '''

    if 'spacing' in information:
        image.SetSpacing(information.pop('spacing'))
    if 'origin' in information:
        image.SetOrigin(information.pop('origin'))
    if 'direction' in information:
        image.SetDirection(information.pop('direction'))

    if 'size' in information:
        assert(np.array_equal( information.pop('size'), image.GetSize() ))
    if 'ncomponents' in information:
        assert( information.pop('ncomponents') == image.GetNumberOfComponentsPerPixel() )
    
    if not len(information) == 0:
        warnings.warn('unwritten keys: {}'.format(','.join(information.keys())))


def fix_array_dimensions(array, ncomponents=1):
    ''' Convenience function that reorders ndarray dimensions for io with SimpleITK

    Parameters
    ----------
    array : np.ndarray
        The array to be reordered
    ncomponents : int, optional
        Number of components per pixel, default 1. 

    Returns
    -------
    np.ndarray : 
        Reordered array

    '''

    act_size = list(array.shape)
    ndims = len(act_size)
    multicomponent = ncomponents > 1

    from_order = list(range( ndims - multicomponent ))
    to_order = list(range( ndims - multicomponent ))[::-1]

    if multicomponent:
        from_order += [-1]
        to_order += [-1]

    return np.ascontiguousarray(np.moveaxis(array, from_order, to_order))


def read_ndarray_with_sitk(path):
    ''' Read a numpy array from a file using SimpleITK

    Parameters
    ----------
    path : str
        Read from this path
    
    Returns
    -------
    image : np.ndarray
        Obtained array
    information : dict
        Additional information about the array

    '''

    image = sitk.ReadImage(str(path))
    information = get_sitk_image_information(image)
    image = sitk.GetArrayFromImage(image)

    image = fix_array_dimensions(image, information['ncomponents'])
    return image, information


def write_ndarray_with_sitk(array, path, **information):
    ''' Write a numpy array to a file using SimpleITK

    Parameters
    ----------
    array : np.ndarray
        Array to be written.
    path : str
        Write to here
    **information : dict
        Contains additional information to be stored in the image file. 
        See set_sitk_image_information for more information.

    '''

    if not 'ncomponents' in information:
        information['ncomponents'] = 1
    ncomponents = information.pop('ncomponents')

    array = fix_array_dimensions(array, ncomponents)
    
    array = sitk.GetImageFromArray(array, ncomponents > 1)
    set_sitk_image_information(array, information)

    sitk.WriteImage(array, str(path))



import logging
import sys
logging.basicConfig(level=logging.DEBUG)

import pytest
import os
import tempfile
import json
from pkg_resources import resource_filename  # @UnresolvedImport
import numpy as np
import pandas as pd

from allensdk.brain_observatory.drifting_gratings import DriftingGratings
from allensdk.brain_observatory.static_gratings import StaticGratings
from allensdk.brain_observatory.natural_movie import NaturalMovie
from allensdk.brain_observatory.natural_scenes import NaturalScenes
from allensdk.brain_observatory.locally_sparse_noise import LocallySparseNoise
from allensdk.brain_observatory.session_analysis import SessionAnalysis
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet as BODS
import allensdk.brain_observatory.stimulus_info as si

if 'TEST_SESSION_ANALYSIS_REGRESSION_DATA' in os.environ:
    data_file = os.environ['TEST_SESSION_ANALYSIS_REGRESSION_DATA']
else:
    data_file = resource_filename(__name__, 'test_session_analysis_regression_data.json')

@pytest.fixture(scope="module")
def paths():
    pyversion = sys.version_info[0]
    logging.debug("loading " + data_file)
    with open(data_file,'r') as f:
        data = json.load(f)
        return data[str(pyversion)]

@pytest.fixture(scope="module")
def nwb_a(paths):
    return paths['nwb_a']

@pytest.fixture(scope="module")
def nwb_b(paths):
    return paths['nwb_b']

@pytest.fixture(scope="module")
def nwb_c(paths):
    return paths['nwb_c']

@pytest.fixture(scope="module")
def analysis_a(paths):
    return paths['analysis_a']

@pytest.fixture(scope="module")
def analysis_b(paths):
    return paths['analysis_b']

@pytest.fixture(scope="module")
def analysis_c(paths):
    return paths['analysis_c']

# session a

@pytest.fixture(scope="module")
def dg(nwb_a, analysis_a):
    return DriftingGratings.from_analysis_file(BODS(nwb_a), analysis_a)

@pytest.fixture(scope="module")
def nm1a(nwb_a, analysis_a):
    return NaturalMovie.from_analysis_file(BODS(nwb_a), analysis_a, si.NATURAL_MOVIE_ONE)

@pytest.fixture(scope="module")
def nm3(nwb_a, analysis_a):
    return NaturalMovie.from_analysis_file(BODS(nwb_a), analysis_a, si.NATURAL_MOVIE_THREE)

# session b

@pytest.fixture(scope="module")
def sg(nwb_b, analysis_b):
    return StaticGratings.from_analysis_file(BODS(nwb_b), analysis_b)

@pytest.fixture(scope="module")
def nm1b(nwb_b, analysis_b):
    return NaturalMovie.from_analysis_file(BODS(nwb_b), analysis_b, si.NATURAL_MOVIE_ONE)

@pytest.fixture(scope="module")
def ns(nwb_b, analysis_b):
    return NaturalScenes.from_analysis_file(BODS(nwb_b), analysis_b)

# session c
@pytest.fixture(scope="module")
def lsn(nwb_c, analysis_c):
    return LocallySparseNoise.from_analysis_file(BODS(nwb_c), analysis_c, si.LOCALLY_SPARSE_NOISE_8DEG)

@pytest.fixture(scope="module")
def nm1c(nwb_c, analysis_c):
    return NaturalMovie.from_analysis_file(BODS(nwb_c), analysis_c, si.NATURAL_MOVIE_ONE)

@pytest.fixture(scope="module")
def nm2(nwb_c, analysis_c):
    return NaturalMovie.from_analysis_file(BODS(nwb_c), analysis_c, si.NATURAL_MOVIE_TWO)

@pytest.fixture(scope="module")
def analysis_a_new(nwb_a):
    with tempfile.NamedTemporaryFile(delete=True) as tf:
        save_path = tf.name

    logging.debug("running analysis a")
    session_analysis = SessionAnalysis(nwb_a, save_path)
    session_analysis.session_a(plot_flag=False, save_flag=True)    
    logging.debug("done running analysis a")
    logging.debug(save_path)

    yield save_path

    if os.path.exists(save_path):
        os.remove(save_path)

@pytest.fixture(scope="module")
def analysis_b_new(nwb_b):
    with tempfile.NamedTemporaryFile(delete=True) as tf:
        save_path = tf.name

    logging.debug("running analysis b")
    session_analysis = SessionAnalysis(nwb_b, save_path)
    session_analysis.session_b(plot_flag=False, save_flag=True)    
    logging.debug("done running analysis b")
    logging.debug(save_path)

    yield save_path

    if os.path.exists(save_path):
        os.remove(save_path)

@pytest.fixture(scope="module")
def analysis_c_new(nwb_c):
    with tempfile.NamedTemporaryFile(delete=True) as tf:
        save_path = tf.name

    logging.debug("running analysis c")
    session_analysis = SessionAnalysis(nwb_c, save_path)

    session_type = BODS(nwb_c).get_metadata()['session_type']
    if session_type == si.THREE_SESSION_C2:
        session_analysis.session_c2(plot_flag=False, save_flag=True)    
    elif session_type == si.THREE_SESSION_C:
        session_analysis.session_c(plot_flag=False, save_flag=True)    
    logging.debug("done running analysis c")
    
    logging.debug(save_path)

    yield save_path

    if os.path.exists(save_path):
        os.remove(save_path)


def compare_peak(p1, p2):
    assert len(set(p1.columns) ^ set(p2.columns)) == 0
    
    p1 = p1.infer_objects()
    p2 = p2.infer_objects()

    peak_blacklist = [ "rf_center_on_x_lsn",
                       "rf_center_on_y_lsn",
                       "rf_center_off_x_lsn",
                       "rf_center_off_y_lsn",
                       "rf_area_on_lsn",
                       "rf_area_off_lsn",
                       "rf_distance_lsn",
                       "rf_overlap_index_lsn",
                       "rf_chi2_lsn" ]

    for col in p1.select_dtypes(include=[np.number]):
        if col in peak_blacklist:
            logging.debug("skipping " + col)
            continue

        logging.debug("checking " + col)
        assert np.allclose(p1[col], p2[col], equal_nan=True)

    for col in p1.select_dtypes(include=['O']):
        logging.debug("checking " + col)
        assert all(p1[col] == p2[col])

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_a(analysis_a, analysis_a_new):
    peak = pd.read_hdf(analysis_a, "analysis/peak")
    new_peak = pd.read_hdf(analysis_a_new, "analysis/peak")
    compare_peak(peak, new_peak)


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_drifting_gratings(dg, nwb_a, analysis_a_new):
    logging.debug("reading outputs")
    dg_new = DriftingGratings.from_analysis_file(BODS(nwb_a), analysis_a_new)
    #assert np.allclose(dg.sweep_response, dg_new.sweep_response)
    assert np.allclose(dg.mean_sweep_response, dg_new.mean_sweep_response, equal_nan=True)
        
    assert np.allclose(dg.response, dg_new.response, equal_nan=True)
    assert np.allclose(dg.noise_correlation, dg_new.noise_correlation, equal_nan=True)
    assert np.allclose(dg.signal_correlation, dg_new.signal_correlation, equal_nan=True)
    assert np.allclose(dg.representational_similarity, dg_new.representational_similarity, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_natural_movie_one_a(nm1a, nwb_a, analysis_a_new):
    nm1a_new = NaturalMovie.from_analysis_file(BODS(nwb_a), analysis_a_new, si.NATURAL_MOVIE_ONE)
    #assert np.allclose(nm1a.sweep_response, nm1a_new.sweep_response)
    assert np.allclose(nm1a.binned_cells_sp, nm1a_new.binned_cells_sp, equal_nan=True)
    assert np.allclose(nm1a.binned_cells_vis, nm1a_new.binned_cells_vis, equal_nan=True)
    assert np.allclose(nm1a.binned_dx_sp, nm1a_new.binned_dx_sp, equal_nan=True)
    assert np.allclose(nm1a.binned_dx_vis, nm1a_new.binned_dx_vis, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")    
def test_natural_movie_three(nm3, nwb_a, analysis_a_new):
    #nm3_new = NaturalMovie.from_analysis_file(BODS(nwb_a), analysis_a_new, si.NATURAL_MOVIE_THREE)
    #assert np.allclose(nm3.sweep_response, nm3_new.sweep_response)
    pass


@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_b(analysis_b, analysis_b_new):
    peak = pd.read_hdf(analysis_b, "analysis/peak")
    new_peak = pd.read_hdf(analysis_b_new, "analysis/peak")
    compare_peak(peak, new_peak)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_static_gratings(sg, nwb_b, analysis_b_new):
    sg_new = StaticGratings.from_analysis_file(BODS(nwb_b), analysis_b_new)
    #assert np.allclose(sg.sweep_response, sg_new.sweep_response)
    assert np.allclose(sg.mean_sweep_response, sg_new.mean_sweep_response, equal_nan=True)

    assert np.allclose(sg.response, sg_new.response, equal_nan=True)
    assert np.allclose(sg.noise_correlation, sg_new.noise_correlation, equal_nan=True)
    assert np.allclose(sg.signal_correlation, sg_new.signal_correlation, equal_nan=True)
    assert np.allclose(sg.representational_similarity, sg_new.representational_similarity, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_natural_movie_one_b(nm1b, nwb_b, analysis_b_new):    
    nm1b_new = NaturalMovie.from_analysis_file(BODS(nwb_b), analysis_b_new, si.NATURAL_MOVIE_ONE)
    #assert np.allclose(nm1b.sweep_response, nm1b_new.sweep_response)

    assert np.allclose(nm1b.binned_cells_sp, nm1b_new.binned_cells_sp, equal_nan=True)
    assert np.allclose(nm1b.binned_cells_vis, nm1b_new.binned_cells_vis, equal_nan=True)
    assert np.allclose(nm1b.binned_dx_sp, nm1b_new.binned_dx_sp, equal_nan=True)
    assert np.allclose(nm1b.binned_dx_vis, nm1b_new.binned_dx_vis, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_natural_scenes(ns, nwb_b, analysis_b_new):    
    ns_new = NaturalScenes.from_analysis_file(BODS(nwb_b), analysis_b_new)
    #assert np.allclose(ns.sweep_response, ns_new.sweep_response)
    assert np.allclose(ns.mean_sweep_response, ns_new.mean_sweep_response, equal_nan=True)

    assert np.allclose(ns.noise_correlation, ns_new.noise_correlation, equal_nan=True)
    assert np.allclose(ns.signal_correlation, ns_new.signal_correlation, equal_nan=True)
    assert np.allclose(ns.representational_similarity, ns_new.representational_similarity, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_session_c(analysis_c, analysis_c_new):
    peak = pd.read_hdf(analysis_c, "analysis/peak")
    new_peak = pd.read_hdf(analysis_c_new, "analysis/peak")
    compare_peak(peak, new_peak)

@pytest.mark.skipif(os.getenv('TEST_COMPETE') != 'true',
                    reason="partial testing")
def test_locally_sparse_noise(lsn, nwb_c, analysis_c_new):
    ds = BODS(nwb_c)
    session_type = ds.get_metadata()['session_type']
    logging.debug(session_type)

    if session_type == si.THREE_SESSION_C:
        lsn_new = LocallySparseNoise.from_analysis_file(ds, analysis_c_new, si.LOCALLY_SPARSE_NOISE)
    elif session_type == si.THREE_SESSION_C2:
        lsn_new = LocallySparseNoise.from_analysis_file(ds, analysis_c_new, si.LOCALLY_SPARSE_NOISE_4DEG)
        
    #assert np.allclose(lsn.sweep_response, lsn_new.sweep_response)
    assert np.allclose(lsn.mean_sweep_response, lsn_new.mean_sweep_response, equal_nan=True)

@pytest.mark.skipif(os.getenv('TEST_COMPLETE') != 'true',
                    reason="partial testing")
def test_natural_movie_one_c(nm1c, nwb_c, analysis_c_new):    
    nm1c_new = NaturalMovie.from_analysis_file(BODS(nwb_c), analysis_c_new, si.NATURAL_MOVIE_ONE)
    #assert np.allclose(nm1c.sweep_response, nm1c_new.sweep_response)

    assert np.allclose(nm1c.binned_dx_sp, nm1c_new.binned_dx_sp, equal_nan=True) 
    assert np.allclose(nm1c.binned_dx_vis, nm1c_new.binned_dx_vis, equal_nan=True) 
    assert np.allclose(nm1c.binned_cells_sp, nm1c_new.binned_cells_sp, equal_nan=True) 
    assert np.allclose(nm1c.binned_cells_vis, nm1c_new.binned_cells_vis, equal_nan=True) 
    
    




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

import functools
import six

import h5py


def decode_bytes(bytes_dataset, encoding='UTF-8'):
    ''' Convert the elements of a dataset of bytes to str
    '''

    return [ item.decode(encoding) for item in bytes_dataset[:].flat ]


def load_datasets_by_relnames(relnames, h5_file, start_node):
    ''' A convenience function for finding and loading into memory one or more
    datasets from an h5 file
    '''

    matcher_cbs = {
        relname: functools.partial(h5_object_matcher_relname_in, [relname]) 
        for relname in relnames
    }

    matches = keyed_locate_h5_objects(matcher_cbs, h5_file, start_node=start_node)
    return { key: value[:] for key, value in six.iteritems(matches) }


def h5_object_matcher_relname_in(relnames, h5_object_name, h5_object):
    ''' Asks if an h5 object's relative name (the final section of its absolute name)
    is contained within a provided array

    Parameters
    ----------
    relnames : array-like
        Relative names against which to match
    h5_object_name : str
        Full name (path from origin) of h5 object
    h5_object : h5py.Group, h5py.Dataset
        Check this object's relative name

    Returns
    -------
    bool : 
        whether the match succeeded
    h5_object : h5py.group, h5py.Dataset
        the argued object

    '''

    return h5_object_name.split('/')[-1] in relnames, h5_object


def keyed_locate_h5_objects(matcher_cbs, h5_file, start_node=None):
    ''' Traverse an h5 file and build up a dictionary mapping supplied keys to 
    located objects
    '''

    matches = {}
    def matcher(obj_name, obj):
        for key, matcher_cb in six.iteritems(matcher_cbs):
            match, _ = matcher_cb(obj_name, obj)
            if match:
                matches[key] = obj

    traverse_h5_file(matcher, h5_file, start_node)
    return matches


def locate_h5_objects(matcher_cb, h5_file, start_node=None):
    ''' Traverse an h5 file and return objects matching supplied criteria
    '''

    matches = []
    def matcher(h5_object_name, h5_object):
        match, _ = matcher_cb(h5_object_name, h5_object)
        if match:
            matches.append(h5_object)

    traverse_h5_file(matcher, h5_file, start_node)
    return matches


def traverse_h5_file(callback, h5_file, start_node=None):
    ''' Traverse an h5 file and apply a callback to each node
    '''

    if start_node is None:
        start_node = h5_file['/']
    elif isinstance(start_node, str):
        start_node = h5_file[start_node]

    start_node.visititems(callback)


# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import functools
import numpy as np
from pkg_resources import resource_filename  # @UnresolvedImport
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet, si
import allensdk.core.brain_observatory_nwb_data_set as bonds
import pytest
import os
import h5py

from allensdk.brain_observatory.brain_observatory_exceptions import MissingStimulusException

from test_h5_utilities import mem_h5


NWB_FLAVORS = []

if 'TEST_NWB_FILES' in os.environ:
    nwb_list_file = os.environ['TEST_NWB_FILES']
else:
    nwb_list_file = resource_filename(__name__, 'nwb_files.txt')

if os.environ.get('TEST_COMPLETE', None) == 'true':
    with open(nwb_list_file, 'r') as f:
        NWB_FLAVORS = [l.strip() for l in f]


@pytest.fixture(params=NWB_FLAVORS)
def data_set(request):
    assert os.path.exists(request.param)
    data_set = BrainObservatoryNwbDataSet(request.param)

    return data_set


@pytest.fixture
def stim_pres_h5(mem_h5):
    def make_stim_pres_h5(stimulus_name):
        mem_h5.create_group('stimulus/presentation/{}'.format(stimulus_name))
        mem_h5.create_group('stimulus/not_presentation/{}'.format(stimulus_name))
        return mem_h5
    return make_stim_pres_h5


@pytest.fixture
def abstract_feature_series_h5(mem_h5):
    def make_abstract_feature_series_h5(stimulus_name, stim_data, features, frame_dur):
        
        stimulus_path = 'stimulus/presentation/{}'.format(stimulus_name)
        frame_dur_path = '{}/frame_duration'.format(stimulus_path)
        features_path = '{}/features'.format(stimulus_path)
        stim_data_path = '{}/data'.format(stimulus_path)

        mem_h5[frame_dur_path] = frame_dur
        mem_h5[stim_data_path] = stim_data
        mem_h5[features_path] = features

        return mem_h5
    return make_abstract_feature_series_h5


def test_acceptance(data_set):
    data_set.get_cell_specimen_ids()
    data_set.get_session_type()
    data_set.get_metadata()
    data_set.get_running_speed()
    data_set.get_motion_correction()


def test_get_roi_ids(data_set):
    ids = data_set.get_roi_ids()
    assert len(ids) == len(data_set.get_cell_specimen_ids())

def test_get_metadata(data_set):
    md = data_set.get_metadata()

    valid_fields = [ 'genotype', 'cre_line', 'imaging_depth_um', 'ophys_experiment_id', 'experiment_container_id',
                     'session_start_time', 'age_days', 'device', 'device_name', 'pipeline_version', 'sex',
                     'targeted_structure', 'excitation_lambda', 'indicator', 'fov', 'session_type', 'specimen_name' ]

    invalid_fields = [ 'imaging_depth', 'age', 'device_string', 'generated_by' ]

    for field in valid_fields:
        assert md[field] is not None

    for field in invalid_fields:
        assert field not in md
    


def test_get_cell_specimen_indices(data_set):
    inds = data_set.get_cell_specimen_indices([])
    assert len(inds) == 0

    ids = data_set.get_cell_specimen_ids()

    inds = data_set.get_cell_specimen_indices(ids)
    assert np.all(np.array(inds) == np.arange(len(inds)))

    inds = data_set.get_cell_specimen_indices([ids[0]])
    assert inds[0] == 0


def test_get_fluorescence_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_fluorescence_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_fluorescence_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_fluorescence_traces([ids[0]])
    assert traces.shape[0] == 1


def test_get_neuropil_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_neuropil_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_neuropil_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_neuropil_traces([ids[0]])
    assert traces.shape[0] == 1


def test_get_dff_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_dff_traces()
    # assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_dff_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_dff_traces([ids[0]])
    assert traces.shape[0] == 1

def test_get_neuropil_r(data_set):

    ids = data_set.get_cell_specimen_ids()
    r = data_set.get_neuropil_r()
    assert len(ids) == len(r)

    r = data_set.get_neuropil_r(ids)
    assert len(ids) == len(r)

    short_list = [ids[0]]
    r = data_set.get_neuropil_r(short_list)
    assert len(short_list) == len(r)

def test_get_corrected_fluorescence_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_corrected_fluorescence_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_corrected_fluorescence_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_corrected_fluorescence_traces([ids[0]])
    assert traces.shape[0] == 1


def test_get_roi_mask(data_set):
    ids = data_set.get_cell_specimen_ids()
    roi_masks = data_set.get_roi_mask()
    assert len(ids) == len(roi_masks)

    max_projection = data_set.get_max_projection()
    for roi_mask in roi_masks:
        mask = roi_mask.get_mask_plane()
        assert mask.shape[0] == max_projection.shape[0]
        assert mask.shape[1] == max_projection.shape[1]

    roi_masks = data_set.get_roi_mask([ids[0]])
    assert len(roi_masks) == 1


def test_get_roi_mask_array(data_set):
    ids = data_set.get_cell_specimen_ids()
    arr = data_set.get_roi_mask_array()
    assert arr.shape[0] == len(ids)

    arr = data_set.get_roi_mask_array([ids[0]])
    assert arr.shape[0] == 1

    try:
        arr = data_set.get_roi_mask_array([0])
    except ValueError as e:
        assert str(e).startswith("Cell specimen not found")


def test_get_stimulus_epoch_table(data_set):

    summary_df = data_set.get_stimulus_epoch_table()

    session_type = data_set.get_session_type()
    if session_type == si.THREE_SESSION_A or si.THREE_SESSION_C:
        assert len(summary_df) == 7
    elif session_type == si.THREE_SESSION_B:
        assert len(summary_df) == 8
    elif session_type == si.THREE_SESSION_C2:
        assert len(summary_df) == 10
    else:
        raise NotImplementedError('Code not tested for session of type: %s' % session_type)

def test_get_stimulus_table_master(data_set):

    master_df = data_set.get_stimulus_table('master')

    session_type = data_set.get_session_type()
    if session_type == si.THREE_SESSION_A:
        assert len(master_df) == 45629
    elif session_type == si.THREE_SESSION_B:
        assert len(master_df) == 20951
    elif session_type == si.THREE_SESSION_C:
        assert len(master_df) == 26882
    elif session_type == si.THREE_SESSION_C2:
        assert len(master_df) == 29398
    else:
        raise NotImplementedError('Code not tested for session of type: %s' % session_type)


def test_make_indexed_time_series_stimulus_table():

    stimulus_name = 'fish'
    frame_dur_exp = np.arange(20).reshape((10, 2))
    inds_exp = np.arange(10)

    obt = bonds._make_indexed_time_series_stimulus_table(inds_exp, frame_dur_exp)

    frame_dur_obt = np.array([ obt['start'].values, obt['end'].values ]).T
    assert(np.allclose( frame_dur_obt, frame_dur_exp ))


def test_make_indexed_time_series_stimulus_table_out_of_order():

    stimulus_name = 'fish'
    frame_dur_exp = np.arange(20).reshape((10, 2))
    frame_dur_file = frame_dur_exp.copy()[::-1, :]
    inds_exp = np.arange(10)

    obt = bonds._make_indexed_time_series_stimulus_table(inds_exp, frame_dur_file)

    frame_dur_obt = np.array([ obt['start'].values, obt['end'].values ]).T
    assert(np.allclose( frame_dur_obt, frame_dur_exp ))


def test_make_abstract_feature_series_stimulus_table_out_of_order():

    stimulus_name = 'fish'
    frame_dur_exp = np.arange(20).reshape((10, 2))
    frame_dur_file = frame_dur_exp.copy()[::-1, :]
    features_exp = ['orientation', 'spatial_frequency', 'phase']
    data_exp = np.arange(30).reshape((10, 3))
    data_file = data_exp.copy()[::-1, :]

    obt = bonds._make_abstract_feature_series_stimulus_table(data_file, features_exp, frame_dur_file)

    frame_dur_obt = np.array([ obt['start'].values, obt['end'].values ]).T
    assert(np.allclose( frame_dur_obt, frame_dur_exp ))

    data_obt = np.array([ obt['orientation'].values, obt['spatial_frequency'].values, obt['phase'].values ]).T
    assert(np.allclose( data_obt, data_exp ))


def test_make_spontanous_activity_stimulus_table():

    table_values_exp = [[0, 2], [4, 6]]

    frame_dur = np.arange(8).reshape((4, 2))
    events = np.array([ 1, -1, 1, -1 ])

    obt = bonds._make_spontaneous_activity_stimulus_table(events, frame_dur)
    assert(np.allclose( obt.values, table_values_exp ))


def test_make_repeated_indexed_time_series_stimulus_table():

    stimulus_name = 'fish'
    frame_dur_exp = np.arange(20).reshape((10, 2))
    inds_exp = np.array([0, 1, 2, 3, 4, 0, 1, 2, 3, 4])
    repeats_exp = np.array([0] * 5 + [1] * 5)
    
    obt = bonds._make_repeated_indexed_time_series_stimulus_table(inds_exp, frame_dur_exp)

    frame_dur_obt = np.array([ obt['start'].values, obt['end'].values ]).T
    assert(np.allclose( frame_dur_obt, frame_dur_exp ))
    assert(np.allclose( repeats_exp, obt['repeat'] ))


def test_find_stimulus_presentation_group(stim_pres_h5):

    stimulus_name = 'fish'
    stim_pres_h5 = stim_pres_h5(stimulus_name)

    obt = bonds._find_stimulus_presentation_group(stim_pres_h5, stimulus_name)

    assert( obt.name == '/stimulus/presentation/fish' )


def test_find_stimulus_presentation_group_missing(stim_pres_h5):

    stimulus_name = 'fish'
    stim_pres_h5 = stim_pres_h5('fowl')

    with pytest.raises(MissingStimulusException):
        obt = bonds._find_stimulus_presentation_group(stim_pres_h5, stimulus_name)


def test_find_stimulus_presentation_group_duplicate(stim_pres_h5):

    stimulus_name = 'fish'
    stim_pres_h5 = stim_pres_h5('fish')
    stim_pres_h5.create_group('/stimulus/presentation/fish_stimulus')

    with pytest.raises(MissingStimulusException):
        obt = bonds._find_stimulus_presentation_group(stim_pres_h5, stimulus_name)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import os
from allensdk.config.manifest_builder import ManifestBuilder
from allensdk.config.manifest import Manifest


@pytest.fixture
def builder():
    b = ManifestBuilder()
    b.add_path('BASEDIR', '/home/username/example')

    return b


def testManifestConstructor(builder):
    manifest = builder.get_manifest()
    expected = os.path.abspath('/home/username/example')
    actual = manifest.get_path('BASEDIR')
    assert(expected == actual)


def testManifestParent(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')
    manifest = builder.get_manifest()
    expected = os.path.abspath('/home/username/example/work')
    actual = manifest.get_path('WORKDIR')
    assert(expected == actual)


def testManifestBuilderDataFrame(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')
    builder_df = builder.as_dataframe()

    assert('key' in builder_df.keys())
    assert('type' in builder_df.keys())
    assert('spec' in builder_df.keys())
    assert('parent_key' in builder_df.keys())
    assert('format' in builder_df.keys())
    assert(5 == len(builder_df.keys()))


def testManifestDataFrame(builder):
    builder.add_path('WORKDIR',
                     'work',
                     parent_key='BASEDIR')

    manifest = builder.get_manifest()
    df = manifest.as_dataframe()

    assert('type' in df.keys())
    assert('spec' in df.keys())
    assert(2 == len(df.keys()))


def safe_mkdir_root_dir():
    directory = os.path.abspath(os.sep)
    Manifest.safe_mkdir(directory) # should not error


# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
import os
import numpy as np
from mock import patch, mock_open, MagicMock
from allensdk.core.brain_observatory_cache import BrainObservatoryCache
from allensdk.api.queries.brain_observatory_api import BrainObservatoryApi
import json
import allensdk.brain_observatory.stimulus_info as si
from allensdk.test_utilities.regression_fixture import get_list_of_path_dict

try:
    import __builtin__ as builtins  # @UnresolvedImport
except:
    import builtins  # @UnresolvedImport


CACHE_MANIFEST = """
{
  "manifest": [
    {
      "type": "manifest_version",
      "value": "1.2"
    },
    {
      "type": "dir",
      "spec": ".",
      "key": "BASEDIR"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "experiment_containers.json",
      "key": "EXPERIMENT_CONTAINERS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_experiments.json",
      "key": "EXPERIMENTS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_experiment_data/%d.nwb",
      "key": "EXPERIMENT_DATA"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "cell_specimens.json",
      "key": "CELL_SPECIMENS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "stimulus_mappings.json",
      "key": "STIMULUS_MAPPINGS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_analysis_data/%d_%s_analysis.h5",
      "key": "ANALYSIS_DATA"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_experiment_events/%d_events.npz",
      "key": "EVENTS_DATA"
    }
  ]
}
"""


@pytest.fixture()
def events_test_data():
    return {"pattern": "/allen/aibs/informatics/module_test_data/observatory/events/%d_events.npz",
            "experiment_id": 715923832}


@pytest.fixture(scope="function")
def brain_observatory_cache():
    boc = None

    try:
        manifest_data = bytes(CACHE_MANIFEST, 'UTF-8')  # Python 3
    except:
        manifest_data = bytes(CACHE_MANIFEST)  # Python 2.7

    with patch('os.path.exists',
               return_value=True):
        with patch(builtins.__name__ + ".open",
                   mock_open(read_data=manifest_data)):
            # Download a list of all targeted areas
            boc = BrainObservatoryCache(manifest_file="some_path/manifest.json",
                                        base_uri='http://api.brain-map.org')

    return boc


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_all_targeted_structures(mock_json_msg_query,
                                     brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                brain_observatory_cache.get_all_targeted_structures()

        mock_json_msg_query.assert_called_once_with(
            "http://api.brain-map.org/api/v2/data/query.json?q="
            "model::ExperimentContainer,rma::include,"
            "ophys_experiments,isi_experiment,"
            "specimen(donor(conditions,age,transgenic_lines)),"
            "targeted_structure,"
            "rma::options[num_rows$eq'all'][count$eqfalse]")


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_experiment_containers(mock_json_msg_query,
                                   brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download experiment containers for VISp experiments
                visp_ecs = brain_observatory_cache.get_experiment_containers(
                    targeted_structures=['VISp'])

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ExperimentContainer,rma::include,"
        "ophys_experiments,isi_experiment,"
        "specimen(donor(conditions,age,transgenic_lines)),targeted_structure,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_all_cre_lines(mock_json_msg_query,
                           brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all cre lines
                tls = brain_observatory_cache.get_all_cre_lines()

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ExperimentContainer,rma::include,"
        "ophys_experiments,isi_experiment,"
        "specimen(donor(conditions,age,transgenic_lines)),targeted_structure,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_ophys_experiments(mock_json_msg_query,
                               brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache.get_ophys_experiments()

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::OphysExperiment,rma::include,experiment_container,"
        "well_known_files(well_known_file_type),targeted_structure,"
        "specimen(donor(age,transgenic_lines)),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_all_session_types(mock_json_msg_query,
                               brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache.get_all_session_types()

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::OphysExperiment,rma::include,experiment_container,"
        "well_known_files(well_known_file_type),targeted_structure,"
        "specimen(donor(age,transgenic_lines)),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_stimulus_mappings(mock_json_msg_query,
                               brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache._get_stimulus_mappings()

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ApiCamStimulusMapping,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@pytest.mark.skipif(True, reason="need to develop mocks")
@patch.object(BrainObservatoryApi, "json_msg_query")
def test_get_cell_specimens(mock_json_msg_query,
                            brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            # Download a list of all transgenic driver lines
            tls = brain_observatory_cache.get_cell_specimens()

    mock_json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q=")


def test_build_manifest(tmpdir_factory):
    try:
        manifest_data = bytes(CACHE_MANIFEST, 'UTF-8')  # Python 3
    except:
        manifest_data = bytes(CACHE_MANIFEST)  # Python 2.7

    manifest_file = str(tmpdir_factory.mktemp("boc").join("manifest.json"))
    with patch('allensdk.config.manifest_builder.ManifestBuilder.write_json_string') as mock_write_json_string:
        mock_write_json_string.return_value = manifest_data

        brain_observatory_cache = BrainObservatoryCache(manifest_file=manifest_file)
        with open(manifest_file, 'rb') as f:
            read_manifest_data = f.read()

        assert manifest_data == read_manifest_data


def test_string_argument_errors(brain_observatory_cache):
    boc = brain_observatory_cache

    with pytest.raises(TypeError):
        boc.get_experiment_containers(targeted_structures='str')

    with pytest.raises(TypeError):
        boc.get_experiment_containers(cre_lines='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(targeted_structures='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(cre_lines='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(stimuli='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(session_types='str')

@pytest.mark.skipif(not os.path.exists('/allen/aibs/informatics/module_test_data'), reason='AIBS path not available')
@pytest.mark.parametrize("path_dict", get_list_of_path_dict())
def test_brain_observatory_cache_get_analysis_file(brain_observatory_cache, path_dict): 

    nwb_path_pattern = os.path.join(os.path.dirname(path_dict['nwb_file']), '%d.nwb') 
    brain_observatory_cache.manifest.add_path(brain_observatory_cache.EXPERIMENT_DATA_KEY, nwb_path_pattern)

    analysis_path_pattern = os.path.join(os.path.dirname(path_dict['analysis_file']), '%d_%s_analysis.h5') 
    brain_observatory_cache.manifest.add_path(brain_observatory_cache.ANALYSIS_DATA_KEY, analysis_path_pattern)

    oeid = path_dict['ophys_experiment_id']
    data_set = brain_observatory_cache.get_ophys_experiment_data(oeid)
    for stimulus in data_set.list_stimuli():
        if stimulus != si.SPONTANEOUS_ACTIVITY:
            brain_observatory_cache.get_ophys_experiment_analysis(oeid, stimulus)


@pytest.mark.skipif(not os.path.exists('/allen/aibs/informatics/module_test_data'), reason='AIBS path not available')
def test_brain_observatory_cache_get_events_data(brain_observatory_cache, events_test_data):
    eid = events_test_data["experiment_id"]
    data_file = events_test_data["pattern"] % eid

    brain_observatory_cache.manifest.add_path(brain_observatory_cache.EVENTS_DATA_KEY, events_test_data["pattern"])

    events = brain_observatory_cache.get_ophys_experiment_events(eid)
    true_events = np.load(data_file, allow_pickle=False)["ev"]
    assert(np.all(events == true_events))



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from .rma_api import RmaApi
from allensdk.api.cache import cacheable, Cache
from allensdk.core.obj_utilities import read_obj
import allensdk.core.sitk_utilities as sitk_utilities
import numpy as np
import nrrd
import six


class ReferenceSpaceApi(RmaApi):

    AVERAGE_TEMPLATE = 'average_template'
    ARA_NISSL = 'ara_nissl'
    MOUSE_2011 = 'annotation/mouse_2011'
    DEVMOUSE_2012 = 'annotation/devmouse_2012'
    CCF_2015 = 'annotation/ccf_2015'
    CCF_2016 = 'annotation/ccf_2016'
    CCF_2017 = 'annotation/ccf_2017'
    CCF_VERSION_DEFAULT = CCF_2017

    VOXEL_RESOLUTION_10_MICRONS = 10
    VOXEL_RESOLUTION_25_MICRONS = 25
    VOXEL_RESOLUTION_50_MICRONS = 50
    VOXEL_RESOLUTION_100_MICRONS = 100


    def __init__(self, base_uri=None):
        super(ReferenceSpaceApi, self).__init__(base_uri=base_uri)


    @cacheable(strategy='create',
               reader=nrrd.read,
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           path_keyword='file_name'))
    def download_annotation_volume(self,
                                   ccf_version,
                                   resolution,
                                   file_name):
        '''
        Download the annotation volume at a particular resolution.

        Parameters
        ----------
        ccf_version: string
            Which reference space version to download. Defaults to "annotation/ccf_2017"
        resolution: int
            Desired resolution to download in microns.
            Must be 10, 25, 50, or 100.
        file_name: string
            Where to save the annotation volume.
        
        Note: the parameters must be used as positional parameters, not keywords
        '''

        if ccf_version is None:
            ccf_version = ReferenceSpaceApi.CCF_VERSION_DEFAULT

        self.download_volumetric_data(ccf_version,
                                      'annotation_%d.nrrd' % resolution, 
                                      save_file_path=file_name)


    @cacheable(strategy='create', reader=sitk_utilities.read_ndarray_with_sitk, 
               pathfinder=Cache.pathfinder(file_name_position=3,
                                           path_keyword='file_name'))
    def download_mouse_atlas_volume(self, age, volume_type, file_name):
        '''Download a reference volume (annotation, grid annotation, atlas volume) 
        from the mouse brain atlas project

        Parameters
        ----------
        age : str
            Specify a mouse age for which to download the reference volume
        volume_type : str
            Specify the type of volume to download
        file_name : str
            Specify the path to the downloaded volume
        '''

        remote_file_name = '{}_{}.zip'.format(age, volume_type)
        url = '/'.join([ self.informatics_archive_endpoint, 
                         'current-release', 'mouse_annotation', 
                         remote_file_name ])

        self.retrieve_file_over_http(url, file_name, zipped=True)


    @cacheable(strategy='create',
               reader=nrrd.read,
               pathfinder=Cache.pathfinder(file_name_position=2,
                                           path_keyword='file_name'))
    def download_template_volume(self, resolution, file_name):
        '''
        Download the registration template volume at a particular resolution.

        Parameters
        ----------

        resolution: int
            Desired resolution to download in microns.  Must be 10, 25, 50, or 100.

        file_name: string
            Where to save the registration template volume.
        '''
        self.download_volumetric_data(ReferenceSpaceApi.AVERAGE_TEMPLATE,
                                      'average_template_%d.nrrd' % resolution, 
                                      save_file_path=file_name)

    @cacheable(strategy='create', 
               reader=nrrd.read, 
               pathfinder=Cache.pathfinder(file_name_position=4, 
                                           path_keyword='file_name'))
    def download_structure_mask(self, structure_id, ccf_version, resolution, file_name):
        '''Download an indicator mask for a specific structure.

        Parameters
        ----------
        structure_id : int
            Unique identifier for the annotated structure
        ccf_version : string
            Which reference space version to download. Defaults to "annotation/ccf_2017"
        resolution : int
            Desired resolution to download in microns.  Must be 10, 25, 50, or 100.
        file_name : string
             Where to save the downloaded mask.

        '''

        if ccf_version  is None:
            ccf_version = ReferenceSpaceApi.CCF_VERSION_DEFAULT

        structure_mask_dir = 'structure_masks_{0}'.format(resolution)
        data_path = '{0}/{1}/{2}'.format(ccf_version, 'structure_masks', structure_mask_dir)        
        remote_file_name = 'structure_{0}.nrrd'.format(structure_id)

        try:
            self.download_volumetric_data(data_path, remote_file_name, save_file_path=file_name)
        except Exception as e:
            self._file_download_log.error('''We weren't able to download a structure mask for structure {0}. 
                                             You can instead build the mask locally using 
                                             ReferenceSpace.many_structure_masks''')
            raise


    @cacheable(strategy='create', 
               reader=read_obj, 
               pathfinder=Cache.pathfinder(file_name_position=3, 
                                           path_keyword='file_name'))
    def download_structure_mesh(self, structure_id, ccf_version, file_name):
        '''Download a Wavefront obj file containing a triangulated 3d mesh built 
        from an annotated structure.

        Parameters
        ----------
        structure_id : int
            Unique identifier for the annotated structure
        ccf_version : string
            Which reference space version to download. Defaults to "annotation/ccf_2017"
        file_name : string
             Where to save the downloaded mask.

        '''

        if ccf_version  is None:
            ccf_version = ReferenceSpaceApi.CCF_VERSION_DEFAULT

        data_path = '{0}/{1}'.format(ccf_version, 'structure_meshes')        
        remote_file_name = '{0}.obj'.format(structure_id)

        try:
            self.download_volumetric_data(data_path, remote_file_name, save_file_path=file_name)
        except Exception as e:
            self._file_download_log.error('unable to download a structure mesh for structure {0}.'.format(structure_id))
            raise


    def build_volumetric_data_download_url(self,
                                           data_path,
                                           file_name,
                                           voxel_resolution=None,
                                           release=None,
                                           coordinate_framework=None):
        '''Construct url to download 3D reference model in NRRD format.

        Parameters
        ----------
        data_path : string
            'average_template', 'ara_nissl', 'annotation/ccf_{year}', 
            'annotation/mouse_2011', or 'annotation/devmouse_2012'
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'

        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_
        for additional documentation.
        '''

        if voxel_resolution is None:
            voxel_resolution = ReferenceSpaceApi.VOXEL_RESOLUTION_10_MICRONS

        if release is None:
            release = 'current-release'

        if coordinate_framework is None:
            coordinate_framework = 'mouse_ccf'

        url = ''.join([self.informatics_archive_endpoint,
                       '/%s/%s/' % (release, coordinate_framework),
                       data_path,
                       '/',
                       file_name])

        return url


    def download_volumetric_data(self,
                                 data_path,
                                 file_name,
                                 voxel_resolution=None,
                                 save_file_path=None,
                                 release=None,
                                 coordinate_framework=None):
        '''Download 3D reference model in NRRD format.

        Parameters
        ----------
        data_path : string
            'average_template', 'ara_nissl', 'annotation/ccf_{year}', 
            'annotation/mouse_2011', or 'annotation/devmouse_2012'
        file_name : string
            server-side file name. 'annotation_10.nrrd' for example.
        voxel_resolution : int
            10, 25, 50 or 100
        coordinate_framework : string
            'mouse_ccf' (default) or 'mouse_annotation'

        Notes
        -----
        See: `3-D Reference Models <http://help.brain-map.org/display/mouseconnectivity/API#API-3DReferenceModels>`_
        for additional documentation.
        '''
        url = self.build_volumetric_data_download_url(data_path,
                                                      file_name,
                                                      voxel_resolution,
                                                      release,
                                                      coordinate_framework)

        if save_file_path is None:
            save_file_path = file_name

        if save_file_path is None:
            save_file_path = 'volumetric_data.nrrd'

        self.retrieve_file_over_http(url, save_file_path)




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.config.manifest_builder import ManifestBuilder
from allensdk.api.cache import Cache
from allensdk.api.queries.reference_space_api import ReferenceSpaceApi
from allensdk.api.queries.ontologies_api import OntologiesApi
from allensdk.deprecated import deprecated

from .ontology import Ontology
from .structure_tree import StructureTree
from .reference_space import ReferenceSpace


class ReferenceSpaceCache(Cache):

    REFERENCE_SPACE_VERSION_KEY = 'REFERENCE_SPACE_VERSION'
    ANNOTATION_KEY = 'ANNOTATION'
    TEMPLATE_KEY = 'TEMPLATE'
    STRUCTURES_KEY = 'STRUCTURES'
    STRUCTURE_TREE_KEY = 'STRUCTURE_TREE'
    STRUCTURE_MASK_KEY = 'STRUCTURE_MASK'
    STRUCTURE_MESH_KEY = 'STRUCTURE_MESH'

    MANIFEST_VERSION = 1.2

    def __init__(self, 
                 resolution, 
                 reference_space_key,
                 **kwargs):

        if not 'version' in kwargs:
            kwargs['version'] = self.MANIFEST_VERSION

        if not 'base_uri' in kwargs:
            kwargs['base_uri'] = None

        super(ReferenceSpaceCache, self).__init__(**kwargs)

        self.resolution = resolution
        self.reference_space_key = reference_space_key        
        
        self.api = ReferenceSpaceApi(base_uri=kwargs['base_uri'])

        
    def get_annotation_volume(self, file_name=None):
        """
        Read the annotation volume.  Download it first if it doesn't exist.

        Parameters
        ----------

        file_name: string
            File name to store the annotation volume.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.

        """

        file_name = self.get_cache_path(
            file_name, self.ANNOTATION_KEY, self.reference_space_key, self.resolution)

        annotation, info = self.api.download_annotation_volume(
            self.reference_space_key,
            self.resolution,
            file_name, 
            strategy='lazy')

        return annotation, info


    def get_template_volume(self, file_name=None):
        """
        Read the template volume.  Download it first if it doesn't exist.

        Parameters
        ----------

        file_name: string
            File name to store the template volume.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.

        """

        file_name = self.get_cache_path(
            file_name, self.TEMPLATE_KEY, self.resolution)

        template, info = self.api.download_template_volume(self.resolution, 
                                                           file_name, 
                                                           strategy='lazy')

        return template, info


    def get_structure_tree(self, file_name=None, structure_graph_id=1):
        """
        Read the list of adult mouse structures and return an StructureTree 
        instance.

        Parameters
        ----------

        file_name: string
            File name to save/read the structures table.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.
        structure_graph_id: int
            Build a tree using structure only from the identified structure graph.
        """
        
        file_name = self.get_cache_path(file_name, self.STRUCTURE_TREE_KEY)

        return OntologiesApi(self.api.api_url).get_structures_with_sets(
            strategy='lazy',
            path=file_name,
            pre=StructureTree.clean_structures,
            post=lambda x: StructureTree(StructureTree.clean_structures(x)), 
            structure_graph_ids=structure_graph_id,
            **Cache.cache_json())


    def get_reference_space(self, structure_file_name=None, 
                            annotation_file_name=None):
        """
        Build a ReferenceSpace from this cache's annotation volume and 
        structure tree. The ReferenceSpace does operations that relate brain 
        structures to spatial domains.
        
        Parameters
        ----------
        
        structure_file_name: string
            File name to save/read the structures table.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.
            
        annotation_file_name: string
            File name to store the annotation volume.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.
        
        """
        
        return ReferenceSpace(self.get_structure_tree(structure_file_name), 
                              self.get_annotation_volume(annotation_file_name)[0], 
                              [self.resolution] * 3)

    def get_structure_mask(self, structure_id, file_name=None, annotation_file_name=None):
        """
        Read a 3D numpy array shaped like the annotation volume that has non-zero values where
        voxels belong to a particular structure.  This will take care of identifying substructures.

        Notes
        -----
        This method downloads structure masks from the Allen Institute. To make your own locally, see 
        ReferenceSpace.many_structure_masks.
        
        Parameters
        ----------

        structure_id: int
            ID of a structure.

        file_name: string
            File name to store the structure mask.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.

        annotation_file_name: string
            File name to store the annotation volume.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.
        """
        structure_id = ReferenceSpaceCache.validate_structure_id(structure_id)

        file_name = self.get_cache_path(
            file_name, self.STRUCTURE_MASK_KEY, self.reference_space_key, 
            self.resolution, structure_id)

        return self.api.download_structure_mask(structure_id, 
                                                self.reference_space_key,
                                                self.resolution,
                                                file_name, 
                                                strategy='lazy')


    def get_structure_mesh(self, structure_id, file_name=None):
        """Obtain a 3D mesh specifying the surface of an annotated structure.
    
        Parameters
        -----------
        structure_id: int
            ID of a structure.
        file_name: string
            File name to store the structure mesh.  If it already exists,
            it will be read from this file.  If file_name is None, the
            file_name will be pulled out of the manifest.  Default is None.

        Returns
        -------
        vertices : np.ndarray
            Dimensions are (nSamples, nCoordinates=3). Locations in the reference space
            of vertices
        vertex_normals : np.ndarray
            Dimensions are (nSample, nElements=3). Vectors normal to vertices.
        face_vertices : np.ndarray
            Dimensions are (sample, nVertices=3). References are given in indices 
            (0-indexed here, but 1-indexed in the file) of vertices that make up each face.
        face_normals : np.ndarray
            Dimensions are (sample, nNormals=3). References are given in indices 
            (0-indexed here, but 1-indexed in the file) of vertex normals that make up each face.

        Notes
        -----
        These meshes are meant for 3D visualization and as such have been smoothed. 
        If you are interested in performing quantative analyses, we recommend that you 
        use the structure masks instead.

        """
        structure_id = ReferenceSpaceCache.validate_structure_id(structure_id)

        file_name = self.get_cache_path(
            file_name, self.STRUCTURE_MESH_KEY, self.reference_space_key, structure_id)

        return self.api.download_structure_mesh(structure_id, 
                                                self.reference_space_key,
                                                file_name, 
                                                strategy='lazy')


    def add_manifest_paths(self, manifest_builder):
        """
        Construct a manifest for this Cache class and save it in a file.

        Parameters
        ----------

        file_name: string
            File location to save the manifest.

        """

        manifest_builder = super(ReferenceSpaceCache, self).add_manifest_paths(manifest_builder)
                                  
        manifest_builder.add_path(self.STRUCTURE_TREE_KEY,
                                  'structures.json',
                                  parent_key='BASEDIR',
                                  typename='file')

        manifest_builder.add_path(self.REFERENCE_SPACE_VERSION_KEY,
                                  '%s',
                                  parent_key='BASEDIR',
                                  typename='dir')

        manifest_builder.add_path(self.ANNOTATION_KEY,
                                  'annotation_%d.nrrd',
                                  parent_key=self.REFERENCE_SPACE_VERSION_KEY,
                                  typename='file')

        manifest_builder.add_path(self.TEMPLATE_KEY,
                                  'average_template_%d.nrrd',
                                  parent_key='BASEDIR',
                                  typename='file')

        manifest_builder.add_path(self.STRUCTURE_MASK_KEY,
                                  'structure_masks/resolution_%d/structure_%d.nrrd',
                                  parent_key=self.REFERENCE_SPACE_VERSION_KEY,
                                  typename='file')

        manifest_builder.add_path(self.STRUCTURE_MESH_KEY,
                                  'structure_meshes/structure_%d.obj',
                                  parent_key=self.REFERENCE_SPACE_VERSION_KEY,
                                  typename='file')

        return manifest_builder


       
 
    @classmethod
    def validate_structure_id(cls, structure_id):

        try:
            structure_id = int(structure_id)
        except ValueError as e:
            raise ValueError("Invalid structure_id (%s): could not convert to integer." % str(structure_id))

        return structure_id


    @classmethod
    def validate_structure_ids(cls, structure_ids):

        for ii, sid in enumerate(structure_ids):
            structure_ids[ii] = cls.validate_structure_id(sid)

        return structure_ids



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2018. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

from mock import MagicMock, patch
import pytest

from allensdk.api.queries.mouse_atlas_api import MouseAtlasApi as MAA



@pytest.fixture
def atlas():
    maa = MAA()
    return maa


@patch.object(MAA, "json_msg_query")
def test_get_genes(mock_query, atlas):

    expected = 'http://api.brain-map.org/api/v2/data/query.json?'\
               'q=model::Gene,rma::criteria,[organism_id$in2],rma::include,chromosome,'\
               'rma::options[num_rows$eq2000][start_row$eq0][order$eq\'id\'][count$eqfalse]'

    for result in atlas.get_genes():
        pass
    
    mock_query.assert_called_once_with(expected)


@patch.object(MAA, "json_msg_query")
def test_get_section_data_sets(mock_query, atlas):

    expected = 'http://api.brain-map.org/api/v2/data/query.json?'\
               'q=model::SectionDataSet,rma::criteria,products[id$in1],rma::include,genes,'\
               'rma::options[num_rows$eq2000][start_row$eq0][order$eq\'id\'][count$eqfalse]'

    for result in atlas.get_section_data_sets():
        pass
    
    mock_query.assert_called_once_with(expected)


def test_download_expression_density(atlas):
    with patch('allensdk.api.api.Api.retrieve_file_over_http') as gda:
        with pytest.raises(RuntimeError):
            atlas.download_expression_density('file.name', 12345)

            gda.assert_called_once_with(
                'http://api.brain-map.org/grid_data/download/'\
                '12345?include=density')
                

def test_download_expression_intensity(atlas):
    with patch('allensdk.api.api.Api.retrieve_file_over_http') as gda:
        with pytest.raises(RuntimeError):
            atlas.download_expression_intensity('file.name', 12345)

            gda.assert_called_once_with(
                'http://api.brain-map.org/grid_data/download/'\
                '12345?include=intensity')


def test_download_expression_energy(atlas):
    with patch('allensdk.api.api.Api.retrieve_file_over_http') as gda:
        with pytest.raises(RuntimeError):
            atlas.download_expression_energy('file.name', 12345)

            gda.assert_called_once_with(
                'http://api.brain-map.org/grid_data/download/'\
                '12345?include=energy')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
from scipy import optimize


class GaussianFitError(RuntimeError): pass


def gaussian2D(height, center_x, center_y, width_x, width_y, rotation):
    '''Build a function which evaluates a scaled 2d gaussian pdf

    Parameters
    ----------
    height : float
        scale factor
    center_x : float
        first coordinate of mean
    center_y : float
        second coordinate of mean
    width_x : float
        standard deviation along x axis
    width_y : float
        standard deviation along y axis
    rotation : float
        degrees clockwise by which to rotate the gaussian

    Returns
    -------
    rotgauss: fn
      parameters are x and y positions (row/column semantics are set by your 
      inputs to this function). Return value is the scaled gaussian pdf 
      evaluated at the argued point.

    '''

    width_x = float(width_x)
    width_y = float(width_y)
    
    rotation = np.deg2rad(rotation)
    center_xp = center_x*np.cos(rotation) - center_y*np.sin(rotation)
    center_yp = center_x*np.sin(rotation) + center_y*np.cos(rotation)
    
    def rotgauss(x,y):
        xp = x*np.cos(rotation) - y*np.sin(rotation)
        yp = x*np.sin(rotation) + y*np.cos(rotation)
        g = height*np.exp(-((center_xp-xp)/width_x)**2/2.0 - ((center_yp-yp)/width_y)**2/2.)
        return g
    return rotgauss
            

def moments2(data):
    '''Treating input image data as an independent multivariate gaussian, 
    estimate mean and standard deviations

    Parameters
    ----------
    data : np.ndarray
        2d numpy array.

    Returns
    -------
    height : float
        The maximum observed value in the data
    y : float
        Mean row index
    x : float
        Mean column index
    width_y : float
        The standard deviation along the mean row 
    width_x : float
        The standard deviation along the mean column
    None : 
        This function returns an instance of None.

    Notes
    -----
    uses original method from website for finding center

    '''

    total = data.sum()

    Y,X = np.indices(data.shape)
    x = ( X * data ).sum() / total
    y = ( Y * data ).sum() / total

    col = data[:, int(np.around(x))]
    width_x = np.sqrt( abs( ( np.arange(col.size) - y ) ** 2 * col ).sum() / col.sum() )

    row = data[int(np.around(y)), :]
    width_y = np.sqrt( abs( ( np.arange(row.size) - x ) ** 2 * row  ).sum() / row.sum() )

    height = data.max()

    return height, y, x, width_y, width_x, None
    

def fitgaussian2D(data):
    '''Fit a 2D gaussian to an image

    Parameters
    ----------
    data : np.ndarray
        input image

    Returns
    -------
    p2 : list
        height
        row mean
        column mean
        row standard deviation
        column standard deviation
        rotation
    
    Notes
    -----
    see gaussian2D for details about output values

    '''

    params = moments2(data)
    def errorfunction(p):
        p2 = np.array([p[0], params[1], params[2], np.abs(p[1]), np.abs(p[2]), p[3]])


        val = np.ravel(gaussian2D(*p2)(*np.indices(data.shape)) - data)

        return (val**2).sum()

    res = optimize.minimize(errorfunction, [ params[0], params[3], params[4], 0.0 ], method='Nelder-Mead', options={'maxfev':2500})
    p = res.x
    p2 = np.array([p[0], params[1], params[2], np.abs(p[1]), np.abs(p[2]), p[3]])
    success = res.success
    if not success and res.status != 2: # Status 2 is loss of precision; might need to handle this separately instead of passing...
        print(success)
        print(res.message)
        print(res.status)
        raise GaussianFitError('Gaussian optimization failed to converge:\n%s' % res.message)

    return p2



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import allensdk.brain_observatory.roi_masks as roi_masks


def test_init_by_pixels():
    a = np.array([[0, 0], [1, 1], [1, 0]])

    m = roi_masks.create_roi_mask(2, 2, [0, 0, 0, 0], pix_list=a)

    mp = m.get_mask_plane()

    assert mp[0, 0] == 1
    assert mp[1, 1] == 1
    assert mp[1, 0] == 0
    assert mp[1, 1] == 1

    assert m.x == 0
    assert m.width == 2
    assert m.y == 0
    assert m.height == 2


def test_init_by_pixels_with_border():
    a = np.array([[1, 1], [2, 1]])

    m = roi_masks.create_roi_mask(3, 3, [1, 1, 1, 1], pix_list=a)

    assert m.x == 1
    assert m.width == 2
    assert m.y == 1
    assert m.height == 1
    assert m.overlaps_motion_border is True


def test_init_by_pixels_large():
    a = np.random.random((512, 512))
    a[a > 0.5] = 1

    m = roi_masks.create_roi_mask(
        512, 512, [0, 0, 0, 0], pix_list=np.argwhere(a))

    npx = len(np.where(a)[0])
    assert npx == len(np.where(m.get_mask_plane())[0])




import os
import sys
import logging
import json
from pkg_resources import resource_filename  # @UnresolvedImport

if 'TEST_SESSION_ANALYSIS_REGRESSION_DATA' in os.environ:
    data_file = os.environ['TEST_SESSION_ANALYSIS_REGRESSION_DATA']
else:
    data_file = resource_filename(__name__, '../test/brain_observatory/test_session_analysis_regression_data_list.json')

def get_list_of_path_dict():
    pyversion = sys.version_info[0]
    logging.debug("loading " + data_file)
    with open(data_file,'r') as f:
        return [curr_fixture for curr_fixture in json.load(f) if curr_fixture['version'] == pyversion]



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
import scipy.interpolate as si
import scipy.ndimage.filters as filt
import scipy.stats as stats


ON_LUMINANCE = 255
OFF_LUMINANCE = 0


def chi_square_binary(events, LSN_template):
    # note: can only be applied to binary events for trial responses
    #
    # *****INPUT*****
    # events: 2D numpy bool with shape (num_trials,num_cells) for presence
    #   or absence of response on a given trial
    # LSN_template: 3D numpy int8 with shape (num_trials,num_y_pixels,num_x_pixels)
    #   for luminance at each pixel location
    #
    # *****OUTPUT*****
    # chi_square_grid_NLL: 3D numpy float with shape (num_cells,num_y_pixels,num_x_pixels)
    #   that gives the p value for the hypothesis that a receptive field is contained
    #   within a 7x7 pixel mask centered on a given pixel location as measured
    #   by a chi-square test for the responses among the pixels (both on and off)
    #   that fall within the mask.

    num_trials = np.shape(events)[0]
    num_cells = np.shape(events)[1]
    num_y = np.shape(LSN_template)[1]
    num_x = np.shape(LSN_template)[2]

    # for each pixel location, get a mask that is centered on that location
    #    disc_masks has shape (num_y,num_x,num_y,num_x)
    disc_masks = get_disc_masks(LSN_template)

    # determine which trials each pixel is active (i.e not gray),
    #   broken up by ON and OFF pixels.
    #   trial_matrix has shape (num_y,num_x,2,num_trials)
    trial_matrix = build_trial_matrix(LSN_template, num_trials)

    # get the total number of trials each pixel is active (i.e. not gray)
    #   trials_per_pixel has shape (num_y,num_x,2)
    trials_per_pixel = np.sum(trial_matrix, axis=3)

    # get the sum of the number of events across all trials that each pixel is active
    #   events_per_pixel has shape (num_cells,num_y,num_x,2)
    events_per_pixel = get_events_per_pixel(events, trial_matrix)

    # smooth stimulus-triggered average spatially with a gaussian
    for n in range(num_cells):
        for on_off in range(2):
            events_per_pixel[n, :, :, on_off] = smooth_STA(events_per_pixel[n, :, :, on_off])

    # calculate the p_value for each exclusion region
    chi_square_grid = np.zeros((num_cells, num_y, num_x))
    for y in range(num_y):
        for x in range(num_x):
            exclusion_mask = np.ones((num_y, num_x, 2)) * disc_masks[y, x, :, :].reshape(num_y, num_x, 1)
            p_vals, __ = chi_square_within_mask(exclusion_mask, events_per_pixel, trials_per_pixel)
            chi_square_grid[:, y, x] = p_vals

    return chi_square_grid


def get_peak_significance(chi_square_grid_NLL,
                          LSN_template,
                          alpha=0.05):
    # ****INPUT*****
    # chi_square_grid_NLL: result of chi_square_binary(events,LSN_template)
    # LSN_template: 3D numpy int8 with shape (num_trials,num_y_pixels,num_x_pixels)
    #   for luminance at each pixel location
    #
    # *****OUTPUT*****
    # significant_cells: 1D numpy bool with shape (num_cells,) that indicates
    #   whether or not each cell has a location with a significant chance of a
    #   true receptive field
    # best_exclusion_region_list: list of 2D numpy bool with len (num_cells). For each cell,
    #   the array is a mask for the best pixel, or all false

    num_cells = np.shape(chi_square_grid_NLL)[0]
    num_y = np.shape(chi_square_grid_NLL)[1]
    num_x = np.shape(chi_square_grid_NLL)[2]

    chi_square_grid = NLL_to_pvalue(chi_square_grid_NLL)

    # get the average size of all masks in units of number of pixels
    disc_masks = get_disc_masks(LSN_template)
    pixels_per_mask_per_pixel = np.sum(disc_masks, axis=(2, 3)).astype(float)

    # find the smallest p-value and determine if it's significant
    significant_cells = np.zeros((num_cells)).astype(bool)
    best_p = np.zeros((num_cells))
    p_value_correction_factor_per_pixel = (1.0 * num_y * num_x / pixels_per_mask_per_pixel)

    best_exclusion_region_list = []
    corrected_p_value_array_list = []
    for n in range(num_cells):

        # Sidak correction:
        p_value_corrected_per_pixel = 1-np.power((1-chi_square_grid[n, :,:]), p_value_correction_factor_per_pixel)
        corrected_p_value_array_list.append(p_value_corrected_per_pixel)

        y, x = np.unravel_index(p_value_corrected_per_pixel.argmin(), (num_y, num_x))

        # if more than one p-value that maxes out, use the median location
        if np.sum(p_value_corrected_per_pixel == 0.0) > 1:

            y, x = np.unravel_index(np.argwhere(p_value_corrected_per_pixel.flatten() == 0.0)[:, 0], (num_y, num_x))
            center_y, center_x = locate_median(y, x)

        best_p[n] = p_value_corrected_per_pixel[y,x]
        if best_p[n] < alpha:
            significant_cells[n] = True
            best_exclusion_region_list.append(disc_masks[y, x, :,:].astype(np.bool))
        else:
            best_exclusion_region_list.append(np.zeros((disc_masks.shape[0], disc_masks.shape[1]), dtype=np.bool))

    return significant_cells, best_p, corrected_p_value_array_list, best_exclusion_region_list


def locate_median(y, x):

    med_x = np.median(x)
    med_y = np.median(y)
    center_x = x[0]
    center_y = y[0]

    for i in range(len(x)):
        dx = x[i] - med_x
        dy = y[i] - med_y
        dc_x = center_x - med_x
        dc_y = center_y - med_y

        if np.sqrt(dx ** 2 + dy ** 2) < np.sqrt(dc_x ** 2 + dc_y ** 2):
            center_x = x[i]
            center_y = y[i]

    return center_y, center_x


def pvalue_to_NLL(p_values, max_NLL=10.0):
    return np.where(p_values == 0.0, max_NLL, -np.log10(p_values))


def NLL_to_pvalue(NLLs, log_base=10.0):
    return (log_base ** (-NLLs))


def get_events_per_pixel(responses_np, trial_matrix):
    '''Obtain a matrix linking cellular responses to pixel activity.

    Parameters
    ----------
    responses_np : np.ndarray
        Dimensions are (nTrials, nCells). Boolean values indicate presence/absence
        of a response on a given trial.
    trial_matrix : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}, nTrials). Boolean values 
        indicate that a pixel was on/off on a particular trial.

    Returns
    -------
    events_per_pixel : np.ndarray
        Dimensions are (nCells, nYPixels, nXPixels, {on, off}). Values for each 
        cell, pixel, and on/off state are the sum of events for that cell across 
        all trials where the pixel was in the on/off state.

    '''

    num_cells = np.shape(responses_np)[1]
    num_y = np.shape(trial_matrix)[0]
    num_x = np.shape(trial_matrix)[1]

    events_per_pixel = np.zeros((num_cells, num_y, num_x, 2))
    for y in range(num_y):
        for x in range(num_x):
            for on_off in range(2):
                frames = np.argwhere(trial_matrix[y, x, on_off, :])[:, 0]
                events_per_pixel[:, y, x, on_off] = np.sum(responses_np[frames, :], axis=0)

    return events_per_pixel


def smooth_STA(STA, gauss_std=0.75, total_degrees=64):
    '''Smooth an image by convolution with a gaussian kernel

    Parameters
    ----------
    STA : np.ndarray
        Input image
    gauss_std : numeric, optional
        Standard deviation of the gaussian kernel. Will be applied to the 
        upsampled image, so units are visual degrees. Default is 0.75
    total_degrees : int, optional
        Size in visual degrees of the input image along its zeroth (row) axis.
        Used to set the scale factor for up/downsampling.

    Returns
    -------
    STA_smoothed : np.ndarray
        Smoothed image
    '''

    deg_per_pnt = total_degrees // STA.shape[0]
    STA_interpolated = interpolate_RF(STA, deg_per_pnt)
    STA_interpolated_smoothed = filt.gaussian_filter(STA_interpolated, gauss_std)
    STA_smoothed = deinterpolate_RF(STA_interpolated_smoothed, STA.shape[1], STA.shape[0], deg_per_pnt)

    return STA_smoothed


def interpolate_RF(rf_map, deg_per_pnt):
    '''Upsample an image
      
    Parameters
    ----------
    rf_map : np.ndarray
        Input image
    deg_per_pnt : numeric
        scale factor

    Returns
    -------
    interpolated : np.ndarray
        Upsampled image
    '''

    x_pnts = np.shape(rf_map)[1]
    y_pnts = np.shape(rf_map)[0]

    x_coor = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, (x_pnts + 1) * deg_per_pnt / 2, deg_per_pnt)
    y_coor = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, (y_pnts + 1) * deg_per_pnt / 2, deg_per_pnt)

    x_interpolated = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (x_pnts / 2 - 1) * deg_per_pnt + 1, 1)
    y_interpolated = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (y_pnts / 2 - 1) * deg_per_pnt + 1, 1)

    interpolated = si.interp2d(x_coor, y_coor, rf_map)
    interpolated = interpolated(x_interpolated, y_interpolated)

    return interpolated


def deinterpolate_RF(rf_map, x_pnts, y_pnts, deg_per_pnt):
    '''Downsample an image

    Parameters
    ----------
    rf_map : np.ndarray
        Input image
    x_pnts : np.ndarray
        Count of sample points along the first (column) axis
    y_pnts : np.ndarray
        Count of sample points along the zeroth (row) axis
    deg_per_pnt : numeric
        scale factor
        
    Returns
    -------
    sampled_yx : np.ndarray
        Downsampled image
    '''

    # x_pnts = 28
    # y_pnts = 16

    x_interpolated = np.arange(-(x_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (x_pnts / 2 - 1) * deg_per_pnt + 1, 1)
    y_interpolated = np.arange(-(y_pnts - 1) * deg_per_pnt / 2, deg_per_pnt / 2 + (y_pnts / 2 - 1) * deg_per_pnt + 1, 1)

    x_deinterpolate = np.arange(0, len(x_interpolated), deg_per_pnt)
    y_deinterpolate = np.arange(0, len(y_interpolated), deg_per_pnt)

    sampled_y = rf_map[y_deinterpolate, :]
    sampled_yx = sampled_y[:, x_deinterpolate]

    return sampled_yx


def chi_square_within_mask(exclusion_mask, events_per_pixel, trials_per_pixel):
    '''Determine if cells respond preferentially to on/off pixels in a mask using
    a chi2 test.
  
    Parameters
    ----------
    exclusion_mask : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}). Integer indicator for INCLUSION (!) 
        of a pixel within the testing region.
    events_per_pixel : np.ndarray
        Dimensions are (nCells, nYPixels, nXPixels, {on, off}). Integer values 
        are response counts by cell to on/off luminance at each pixel.
    trials_per_pixel : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}). Integer values are 
        counts of trials where a pixel is on/off.

    Returns
    -------
    p_vals : np.ndarray 
        One-dimensional, of length nCells. Float values are p-values 
        for the hypothesis that a given cell has a receptive field within the 
        exclusion mask.
    chi : np.ndarray
        Dimensions are (nCells, nYPixels, nXPixels, {on, off}). Values (float) 
        are squared residual event counts divided by expected event counts.
    '''

    num_y = np.shape(exclusion_mask)[0]
    num_x = np.shape(exclusion_mask)[1]

    # d.f. is number of pixels in mask minus one
    degrees_of_freedom = int(np.sum(exclusion_mask)) - 1

    #   observed_by_pixel has shape (num_cells,num_y,num_x,2)
    expected_by_pixel = get_expected_events_by_pixel(exclusion_mask, events_per_pixel, trials_per_pixel)
    observed_by_pixel = (events_per_pixel * exclusion_mask.reshape(1, num_y, num_x, 2)).astype(float)

    # calculate test statistic given observed and expected
    residual_by_pixel = observed_by_pixel - expected_by_pixel
    chi = (residual_by_pixel ** 2) / expected_by_pixel
    chi_sum = np.nansum(chi, axis=(1, 2, 3))

    # get p-value given test statistic and degrees of freedom
    p_vals = 1.0 - stats.chi2.cdf(chi_sum, degrees_of_freedom)

    return p_vals, chi


def get_expected_events_by_pixel(exclusion_mask, events_per_pixel, trials_per_pixel):
    '''Calculate expected number of events per pixel

    Parameters
    ----------
    exclusion_mask : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}). Integer indicator for INCLUSION (!) 
        of a pixel within the testing region.
    events_per_pixel : np.ndarray
        Dimensions are (nCells, nYPixels, nXPixels, {on, off}). Integer values 
        are response counts by cell to on/off luminance at each pixel.
    trials_per_pixel : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}). Integer values are 
        counts of trials where a pixel is on/off.

    Returns
    -------
    np.ndarray :
        Dimensions (nCells, nYPixels, nXPixels, {on, off}). Float values are 
        pixelwise counts of events expected if events are evenly distributed
        in mask across trials.
    '''

    num_y = np.shape(exclusion_mask)[0]
    num_x = np.shape(exclusion_mask)[1]
    num_cells = np.shape(events_per_pixel)[0]

    exclusion_mask = exclusion_mask.reshape(1, num_y, num_x, 2)
    trials_per_pixel = trials_per_pixel.reshape(1, num_y, num_x, 2)

    masked_trials = exclusion_mask * trials_per_pixel
    masked_events = exclusion_mask * events_per_pixel

    total_trials = np.sum(masked_trials).astype(float)
    total_events_by_cell = np.sum(masked_events, axis=(1, 2, 3)).astype(float)
    
    expected_by_cell_per_trial = total_events_by_cell / total_trials
    return masked_trials * expected_by_cell_per_trial.reshape(num_cells, 1, 1, 1)


def build_trial_matrix(LSN_template, 
                       num_trials, 
                       on_off_luminance=(ON_LUMINANCE, OFF_LUMINANCE)):
    '''Construct indicator arrays for on/off pixels across trials.

    Parameters
    ----------
    LSN_template : np.ndarray
        Dimensions are (nTrials, nYPixels, nXPixels). Luminance values per pixel 
        and trial. The size of the first dimension may be larger than the num_trials 
        argument (in which case only the first num_trials slices will be used) 
        but may not be smaller.
    num_trials : int
        The number of trials (left-justified) to build indicators for.
    on_off_luminance : array-like, optional
        The zeroth element is the luminance value of a pixel when on, the first when off.
        Defaults are [255, 0].

    Returns
    -------
    trial_mat : np.ndarray
        Dimensions are (nYPixels, nXPixels, {on, off}, nTrials). Boolean values 
        indicate that a pixel was on/off on a particular trial.
    '''

    _, num_y, num_x = np.shape(LSN_template)
    trial_mat = np.zeros( (num_y, num_x, 2, num_trials), dtype=bool )

    for y in range(num_y):
        for x in range(num_x):
            for oo, on_off in enumerate(on_off_luminance):

                frame = np.argwhere( LSN_template[:num_trials, y, x] == on_off )[:, 0]
                trial_mat[y, x, oo, frame] = True

    return trial_mat


def get_disc_masks(LSN_template, radius=3, on_luminance=ON_LUMINANCE, off_luminance=OFF_LUMINANCE):
    '''Obtain an indicator mask surrounding each pixel. The mask is a square, excluding pixels which 
    are coactive on any trial with the main pixel.

    Parameters
    ----------
    LSN_template : np.ndarray
        Dimensions are (nTrials, nYPixels, nXPixels). Luminance values per pixel 
        and trial.
    radius : int
        The base mask will be a box whose sides are 2 * radius + 1 in length.
    on_luminance : int, optional
        The value of the luminance for on trials. Default is 255
    off_luminance : int, optional
        The value of the luminance for off trials. Default is 0
  

    Returns
    -------
    masks : np.ndarray
        Dimensions are (nYPixels, nXPixels, nYPixels, nXPixels). The first 2 
        dimensions describe the pixel from which the mask was computed. The last 
        2 serve as the dimensions of the mask images themselves. Masks are binary 
        arrays of type float, with 1 indicating inside, 0 outside.
    '''

    num_y = np.shape(LSN_template)[1]
    num_x = np.shape(LSN_template)[2]

    # convert template to true on trials a pixel is not gray and false when gray
    LSN_binary = np.where(LSN_template == off_luminance, 1, LSN_template)
    LSN_binary = np.where(LSN_binary == on_luminance, 1, LSN_binary)
    LSN_binary = np.where(LSN_binary == 1, 1.0, 0.0)

    # get number of trials each pixel is not gray
    on_trials = LSN_binary.sum(axis=0).astype(float)  # shape is (num_y,num_x)

    masks = np.zeros((num_y, num_x, num_y, num_x))
    for y in range(num_y):
        for x in range(num_x):
            trials_not_gray = np.argwhere( LSN_binary[:, y, x] > 0 )[:, 0]
            raw_mask = np.divide( LSN_binary[trials_not_gray, :, :].sum(axis=0), on_trials ) 

            center_y, center_x = np.unravel_index( raw_mask.argmax(), (num_y, num_x) )

            # include center pixel in mask
            raw_mask[center_y, center_x] = 0.0

            x_max = center_x + radius + 1
            if x_max > num_x:
                x_max = num_x
            x_min = center_x - radius
            if x_min < 0:
                x_min = 0
            y_max = center_y + radius + 1
            if y_max > num_y:
                y_max = num_y
            y_min = center_y - radius
            if y_min < 0:
                y_min = 0

            # don't include far away pixels that just happen
            # to not have any trials in common with center pixel
            clean_mask = np.ones(np.shape(raw_mask))
            clean_mask[y_min:y_max, x_min:x_max] = raw_mask[y_min:y_max, x_min:x_max]

            masks[y, x, :, :] = clean_mask

    masks = np.where(masks > 0, 0.0, 1.0)

    return masks

